﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [StorableType("4B69A82A-265B-46DA-9055-B6E0EB6C3EC8")]
  public class SymbolicRegressionSingleObjectiveMetaModelAnalyzer 
    : SymbolicRegressionMetaModelAnalyzer<SymbolicRegressionSingleObjectiveProblem>, ISymbolicExpressionTreeAnalyzer {

    [StorableConstructor]
    protected SymbolicRegressionSingleObjectiveMetaModelAnalyzer(StorableConstructorFlag _) : base(_) { }

    protected SymbolicRegressionSingleObjectiveMetaModelAnalyzer(SymbolicRegressionSingleObjectiveMetaModelAnalyzer original, Cloner cloner) : base(original, cloner) { }

    public SymbolicRegressionSingleObjectiveMetaModelAnalyzer() { }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() { }

    public override IDeepCloneable Clone(Cloner cloner) => new SymbolicRegressionSingleObjectiveMetaModelAnalyzer(this, cloner);

    protected override void PerformApply(SymbolicRegressionSingleObjectiveProblem problem, string targetVariable) {
      double bestQuality = problem.Maximization.Value ? double.MinValue : double.MaxValue;
      SymbolicRegressionSolution bestMetaModel = null;
      foreach (var tree in this.SymbolicExpressionTree.ToArray()) {
        IDataset dataset = problem.ProblemData.Dataset;
        IEnumerable<int> rows = Enumerable.Range(0, dataset.Rows);
        var quality = problem.Evaluator.Evaluate(ExecutionContext, tree, problem.ProblemData, rows);
        
        bool isBetter = problem.Maximization.Value ? (bestQuality < quality) : (bestQuality > quality);
        if(isBetter) {
          bestQuality = quality;
          var model = new SymbolicRegressionModel(
            targetVariable, 
            (ISymbolicExpressionTree)tree.Clone(), 
            new SymbolicDataAnalysisExpressionTreeInterpreter());
          bestMetaModel = new SymbolicRegressionSolution(model, problem.ProblemData);
        }
      }
      BestMetaModelParameter.ActualValue = bestMetaModel;
    }
  }
}
